#!/usr/local/bin/perl -w

=head1 NAME

runtests.pl

=head1 SYNOPSIS

This is a script which allows running various experiments using different features extracted using NSP and the machine learning algorithms of Weka. 

=head1 DESCRIPTION

runtests.pl is currently only a "developer mode" script. Customizing experiments is possible only by editing the Perl code.

By default this script runs the Weka Support Vector Machine classifier with ten fold cross validation, on unigram and bigram features extracted with different selection criteria.

The recommended steps to speed up experiments are:

1. Modify the current script to identify the features that are required.

2. Run the script once with only one Weka classifier, using copies of test collection made by the program utils/createcopies.pl.

3. Create multiple copies of these set of test directories from the first classifier using the program utils/copytestdirs.pl. This saves performing identical feature identification for multiple classifiers.

4. Now change the call to wsdshelldriver.pl in this script to just execute the final classification step using the "--step 6" option, and execute the script with multiple classifiers now added. Or,

5. Create a copies of the script modified in (4) above for each Weka classifier that you wish to test and run these copies simultaneously if multiple computing resources are available.

=head1 USAGE

See "runtests.pl --help" for a list of options and their usage.

=head1 AUTHOR

Mahesh Joshi <joshi031@d.umn.edu>

Ted Pedersen <tpederse@d.umn.edu>

=head1 BUGS

=head1 SEE ALSO

wsdshell.pl

=head1 COPYRIGHT

Copyright (C) 2005, Ted Pedersen and Mahesh Joshi

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=cut

# Check for the first mandatory argument
if (@ARGV != 1) {

    print "ERROR: Missing argument.\n";
    print "USAGE: runtests.pl TEST_DIRECTORY_START_INDEX\n";
    print "See `runtests.pl --help` for options or `perldoc runtests.pl` for details\n";
    exit(1);
}

if (defined $ARGV[0] && $ARGV[0] eq "--help") {

    print "runtests.pl: Perl script to execute a set of Word Sense\n";
    print "Disambiguation experiments using various NSP parameters\n";
    print "and Weka classifiers.\n\n";

    print "USAGE: runtests.pl TEST_DIRECTORY_START_INDEX\n\n";
 
    print "NOTE: This is currently a \"developer mode\" script. Refer to\n";
    print "`perldoc runtests.pl` for details.\n\n";

    print "e.g. runtests.pl 1\n\n";

    print "will begin WSD experiments starting from directory \"test1\"\n";
    print "and continue upto a testX directory depending upon the Perl code.\n";

    exit(0);
}

# Grab the first test directory number
$index=$ARGV[0];

$runtimepath = $0;
if ($runtimepath =~ /(.*\/)/) {

    # Set the path to required files using the path from
    # which the script is called
    $tokenfile = "$1../params/tokens";
    $nontokenfile = "$1../params/nontokens";
    $stopfile = "$1../params/stoplist";

    print $tokenfile . "\n";

} else {

    # Assume the required files are in the current directory
    $tokenfile = "./tokens";
    $nontokenfile = "./nontokens";
    $stopfile = "./stoplist";
}

# Topmost loop should be modified for adding or removing Weka classifiers
foreach $wekamethod ("weka.classifiers.functions.SMO") {

    # Select the various Ngram values to be used
    foreach $ngram (1, 2) {

        # Select the various frequency cut-off values
        foreach $countopts ("--remove 2", "--remove 3", "--remove 4", "--remove 5") {

            # Select the different Weka options
            foreach $wekaopts ("-i") {

                if ($ngram > 1) {

                    # For Ngram > 1, choose the window size
					foreach $countopts2 ("--window 2", "--window 3", "--window 4", "--window 5") {

                    # In addition to frequency cutoff, for Ngrams > 1 select
                    # the various Log Likelihood cutoff values - assumes
                    # use of log likelihood measure for significance testing.
                    # There are many others in the NSP package.
					foreach $statopts ("--score 3.841", "--score 6.635") {

                        	$testdir = "test$index";

                        	if (-e $testdir) {

                            	print STDERR "Processing directory: $testdir\n";

                                # Prepare parameters for wsdshelldriver.pl
                                # Here is where one should add the "--step 6"
                                # option if required
                            	@params = ($testdir, "--wekaclassifier", $wekamethod, "--ngram", $ngram, "--countopts", "$countopts $countopts2", "--wekaopts", $wekaopts, "--statopts", $statopts, "--token", $tokenfile, "--nontoken", $nontokenfile, "--stop", $stopfile, "--prepopts", "--removeNotToken", "--verbose");
                            
                            	system("wsdshelldriver.pl", @params); 

                            	print STDERR "Completed processing directory: $testdir\n";

                            	$index = $index + 1;

                        	} else {

                            	print STDERR "ERROR: Directory '$testdir' does not exist! Aborting..\n";
                            	exit(1);
                        	}
                    	}
					}
	
                } else {

                    # This processing is for unigrams
                    $testdir = "test$index";

                    if (-e $testdir) {

                        print STDERR "Processing directory: $testdir\n";

                        # Prepare parameters for wsdshelldriver.pl
                        # Here is where one should add the "--step 6"
                        # option if required
                        @params = ($testdir, "--wekaclassifier", $wekamethod, "--ngram", $ngram, "--countopts", $countopts, "--wekaopts", $wekaopts, "--token", $tokenfile, "--nontoken", $nontokenfile, "--stop", $stopfile, "--prepopts", "--removeNotToken", "--verbose");

                        system("wsdshelldriver.pl", @params); 

                        print STDERR "Completed processing directory: $testdir\n";

                        $index = $index + 1;

                    } else {

                        print STDERR "ERROR: Directory '$testdir' does not exist! Aborting..\n";
                        exit(1);
                    }
                }
            }
        }
    }
}

