##############
# fatalError #
##############

# This function should be called when a critical error has occurred and the program cannot proceed anymore. It prints out the specified message at STDERR and exits from the program.

# The first parameter to it indicates whether or not the caller wants this function to print out the program usage.

# Any subsequent parameters are treated as message to be printed. In case of just one additional parameter, we assume that a simple print is required by the caller. In case of multiple subsequent arguments, we assume that a formatted printing is required and we invoke printf

# A newline is added at the end of the message by default
sub fatalError {

    # Check if showUsage should be printed after the error or not
    my $show_usage = shift;

    # Print out the message depending upon the number of additional parameters
    print STDERR "ERROR: ";

    if (@_ > 1) {
        printf STDERR @_;
    } else {
        print STDERR $_[0];
    }

    print STDERR "\n";

    # Print out usage if required
    if ($show_usage) {
        showUsage();
    }

    # Terminate the program, since this is a critical error
    exit(-1);
}

############
# tr_print #
############

# This function prints out the specified trace message at STDERR only when the tracing mode is on

# A single argument is treated as a plain string to be printed using print.
# Multiple arguments are treated as a request for formatted print, with the first argument being the format string.

# A newline is added at the end of the message by default
sub tr_print {

    if ($trace == 1) {

        if (@_ > 1) {
            printf STDERR @_;
        } else {
            print STDERR $_[0];
        }

        print STDERR "\n";
    }
}

############
# vb_print #
############

# This function prints out the specified verbose message at STDERR only when the verbose mode is on

# A single argument is treated as a plain string to be printed using print.
# Multiple arguments are treated as a request for formatted print, with the first argument being the format string.

# A newline is added at the end of the message by default
sub vb_print {

    if ($verbose == 1) {

        if (@_ > 1) {
            printf STDERR @_;
        } else {
            print STDERR $_[0];
        }

        print STDERR "\n";
    }
}

############
# sysCall #
############

# This function simply makes a system function call. It adds an error checking functionality and exits from the program if the system call has returned a non-zero exit status for the command that was executed.

sub sysCall {

    tr_print("About to execute: @_");

    my $status = system(@_);

    # Error handling below taken from documentation of system call from 'perldoc perlfunc'
    unless ($status == 0) {

        if ($? == -1) {

            fatalError(0, "Failed to execute \"@_\": $!");

        } elsif ($? & 127) {

            fatalError(0, "Failed while executing \"@_\". Process died with signal %d, %s coredump",
                     ($? & 127),  ($? & 128) ? 'with' : 'without');
        } else {

            fatalError(0, "Failed while executing \"@_\". Process exited with value %d", $? >> 8);
        }
    }
}


