#!/usr/local/bin/perl -w

=head1 NAME

wsdsummary.pl

=head1 SYNOPSIS

This is a program to extract summary information for multiple WSD experiments involving different features and Weka classifiers. 

=head1 DESCRIPTION

See "wsdsummary.pl --help" for a list of options and their usage.

=head1 USAGE

See "wsdsummary.pl --help" for a list of options and their usage.

=head1 AUTHOR

Mahesh Joshi <joshi031@d.umn.edu>

Ted Pedersen <tpederse@d.umn.edu>

=head1 BUGS

=head1 SEE ALSO

wsdshell.pl

=head1 COPYRIGHT

Copyright (C) 2005, Ted Pedersen and Mahesh Joshi

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=cut


use strict;

# Print help if asked for
if (defined $ARGV[0] && $ARGV[0] eq "--help") {

    print "wsdsummary.pl: Program to extract the summary of results for WSD\n";
    print "experiments.\n\n";
    
    print "USAGE: wsdsummary.pl ROOT_DIRECTORY [TOLERANCE] [THRESHOLD] " .
    "[TOPXCLASSES]\n";
   
    print "The root directory argument is mandatory. It specifies the top\n";
    print "level directory that contains the individual testX directories.\n\n";
    
    print "TOLERANCE specifies the criteria for selecting top classifiers.\n";
    print "A value of 3 means that if the best classifier has 75 percent\n";
    print "accuracy, then any other classifier with accuracy between 72-74\n";
    print "is also considered among the best. Default is 3.\n\n";

    print "THRESHOLD specifies the minimum improvement that a classifier\n";
    print "should have over the majority classifier, to be considered\n";
    print "significant. Default is 5.\n\n";

    print "TOPXCLASSES specifies What number of top classifiers should\n";
    print "be counted as best classifiers for counting. This overrides the\n";
    print "tolerance value - to facilitate finding just the best, or top 2\n";
    print "best classifiers etc. The default is same as (tolerance+1),\n";
    print "maps to a range of 72-75 in above example.\n\n";

    print "IMPORTANT NOTE: For wsdsummary.pl to work correctly, the results\n";
    print "of the majority classifer should be necessarily in the \"test0\"\n";
    print "directory inside the root directory.\n";
    
    exit(0);
}

# Get the first parameter as the root directory containing the results
my $root_dir = shift @ARGV;

if (!defined $root_dir) {

    print "Missing arguments!\n";
    print "USAGE: wsdsummary.pl ROOT_DIRECTORY [TOLERANCE] [THRESHOLD] " .
    "[TOPXCLASSES]\n";
    print "See 'wsdsummary --help' for details.\n";
    
    exit(0);
}

# Set default values for tolerance, threshold and topx
my $tolerance = 3;
my $threshold = 5;
my $usetopx = $tolerance + 1;

# Take user defined values for tolerance, threshold and topx if
# specified and if they are valid
if (defined $ARGV[0] && $ARGV[0] > 0) {

    $tolerance = shift @ARGV;
}

if (defined $ARGV[0] && $ARGV[0] > 0) {

    $threshold = shift @ARGV;
}

if (defined $ARGV[0] && $ARGV[0] > 0) {

    $usetopx = shift @ARGV;
}

print STDERR "Root directory: $root_dir\n";

# Open the root directory for reading
opendir (HROOT, $root_dir) or die "Failed to open specified directory $root_dir";

# This contains the contents of the options.log file for each
# of the test case
my %options_hash;

# This contains the mapping of the test directory name to the
# WEKA classifier that was run in that directory
my %testtoclassifier;

# This is a really complex hash containing the entire set of
# results read from the WEKA output file
my %results;

# Variable to hold each test directory name as we process it
my $test_dir;

# Hold the word directory name (also the word) here
my $word_dir;
 
# Start reading the directory in until we exhaust its contents
while ($test_dir = readdir HROOT) {

    print STDERR "Obtained from directory contents: $test_dir\n";

    # Got the directory entry
    # Process only if it is not a "." or a ".." and it is a directory,
    # not a file or anything else
    if ($test_dir !~ /^\..*/ && -d "$root_dir/$test_dir") {

        print STDERR "Processing directory: $test_dir\n";

        # Open the options file in the test directory, bail out if
        # we do not find one.
        open (HOPTIONS, "$root_dir/$test_dir/options.log") 
            or die "Failed to open option specification file " . 
            "$root_dir/$test_dir/options.log";

        # We want to read in the entire file, save old record seperator
        # and undefine it
        my $oldsep = $/;
        undef $/;

        # Read in the entire file contents and shove them away
        # in the options hash
        $options_hash{$test_dir} = <HOPTIONS>;

        # Find the line that specifies the WEKA classifier used
        # and extract its name
        if ($options_hash{$test_dir} =~ /WEKA CLASSIFIER: (.*?)\n/) {

            # Put it into the test to classifier hash
            $testtoclassifier{$test_dir} = $1;

        } else {

            # If not found, then it means that the default was used
            # This is the SMO
            $testtoclassifier{$test_dir} = "weka.classifiers.functions.SMO";
        }

        # Close the file and restore the old record seperator
        close HOPTIONS;
        $/ = $oldsep;

        # Now open the actual test directory to process each word
        opendir (HTEST, "$root_dir/$test_dir") or die "Failed to open test directory $root_dir/$test_dir";

       # Start processing the directory contents
        while ($word_dir = readdir HTEST) {

            print STDERR "Obtained from directory contents: $word_dir\n";

            # Again, skip ., .. and non-directories
            if ($word_dir !~ /^\..*/ && -d "$root_dir/$test_dir/$word_dir") {

                print STDERR "Processing directory: $word_dir\n";

                # We expect two kinds of results in each word directory
                # An abstract result and a sentence result
                my $file = "$root_dir/$test_dir/$word_dir/$word_dir" . 
                ".abstract.sval2.wekaoutput";

                # Process only if its a file
                if (-f $file) {

                    # Read results from the file and fill in the global
                    # hash

                    # Second parameter indicates 0 = abstract and 1 = sentence
                    readResults($file, 0) == 0 
                        or die "Failed to read results for abstract.\n";
                } else {

                    $results{$word_dir}{$test_dir}{0} = -1;
                    readSenses($file, 0) == 0 
                        or die "Failed to read senses for abstract.\n";
                }

                $file = "$root_dir/$test_dir/$word_dir/$word_dir" . 
                ".sentence.sval2.wekaoutput";

                # Process only if its a file
                if (-f $file) {

                    # Read results from the file and fill in the global
                    # hash

                    # Second parameter indicates 0 = abstract and 1 = sentence
                    readResults($file, 1) == 0 
                        or die "Failed to read results for sentence.\n";
                } else {

                    $results{$word_dir}{$test_dir}{1} = -1;
                    readSenses($file, 1) == 0 
                        or die "Failed to find senses for sentence.\n";
                }
            }
        }
    }
} 

my $key;
my $val;

# Now print out the legend which specifies what test case number
# maps to which classifier and the various options used with them
my $legend;
my $headrow = sprintf("%20s", "Word");

foreach $key (sort keys %options_hash) {

    # $legend contains the options
    $legend .= "$key => $options_hash{$key}\n\n";

    # $headrow contains the entire string of test
    # directory names, to be printed for abstrac as well
    # as sentence results
    $headrow .= sprintf("%10s", $key);
}

print "LEGEND:\n\n$legend\n\n";

# Now we want to print the results
# These will be separate for abstract and sentence context
# and then some combined global results
print "RESULTS\n";
print "=======\n\n";

print "ABSTRACT CONTEXT\n";
print "================\n\n";

print "$headrow\n\n";
print "=" x length($headrow);
print "\n\n";

printResults(0) == 0 
    or die "Failed to show abstract results.\n";

print "\n\nSENTENCE CONTEXT\n";
print "================\n\n";

print "$headrow\n\n";
print "=" x length($headrow);
print "\n\n";

printResults(1) == 0 
    or die "Failed to show abstract results.\n";

# Now we try to find the best results for each word

# Gather statistics at the level of a specific test and its parameters
# We generate classifier statistics at abstract level, sentence level and
# at global level. These are stored in the hashes:
# %teststats{0}, %teststats{1} and %teststats{2} respectively
my %teststats;

# We also gather statistics at the level of a classifier (without considering
# any specific options)
# These are stored in %classifierstats
my %classifierstats;

# A temporary hash is populated before updating %classifierstats
my %classifierstatstemp;

# We then maintain performance numbers for each classifier
my %perfstats;

# This array will contain the list of words that were skipped for abstract
# analysis - and should therefore be skipped for sentence analysis
my @words_skipped;


print "\n\nBEST CASE CLASSIFIERS\n";
print "=======================\n\n";

print "ALL FOLLOWING BEST CLASSIFIERS ARE LISTED ONLY IF THEIR PERFORMANCE IS AT LEAST $threshold PERCENT BETTER THAN THE MAJORITY CLASSIFIER.\n\n";

# Initially, print the accuracy figures for the top classifiers

# First for abstract context
print "ABSTRACT: The best classifiers (and top $tolerance% below them) are:\n";
print "===================================================================\n\n";

printf "%20s%60s\n", "Word", "Best classifiers";
print "=" x 100 . "\n";

# Call findBest with abstract context
findBest(0);

# Then for sentence context
print "SENTENCE: The best classifiers (and top $tolerance% below them) are:\n";
print "===================================================================\n\n";

printf "%20s%60s\n", "Word", "Best classifiers";
print "=" x 100 . "\n";

# Call findBest with sentence context
findBest(1);

# Then we want to print the class level statistics
print "\n\nBEST CASE SUMMARY\n";
print "===================\n\n";

# For abstract
print "BEST CLASSIFIER STATISTICS - ABSTRACT RESULTS (ONLY TOP $usetopx BEST COUNTED)\n";
print "==============================================================================\n";
printf "%40s%15s", "Classifier", "Best for #";
print "     Best for words\n";
print "=" x 100 . "\n";

# Call printClassStats with abstract context
printClassStats(0);

# For sentence
print "\n\n\nBEST CLASSIFIER STATISTICS - SENTENCE RESULTS (ONLY TOP $usetopx BEST COUNTED)\n";
print "====================================================================================\n";
printf "%40s%15s", "Classifier", "Best for #";
print "     Best for words\n";
print "=" x 100 . "\n";

# Call printClassStats with sentence context
printClassStats(1);

# Globally
print "\n\n\nBEST CLASSIFIER STATISTICS - GLOBAL RESULTS (ONLY TOP $usetopx BEST COUNTED)\n";
print "==================================================================================\n";
printf "%40s%15s", "Classifier", "Best for #";
print "     Best for words\n";
print "=" x 100 . "\n";

# Call printClassStats with global context
printClassStats(2);

print "\n\n\n";

# Now we have to print the more detailed test level statistics

# For abstract
print "\n\nBEST TEST STATISTICS - ABSTRACT RESULTS\n";
print "==========================================\n\n";
printf "%40s%15s", "Classifier", "Best for #";
print "     Best for words\n";
print "=" x 100 . "\n";

# Call printTestStats with abstract context
printTestStats(0);

# For sentence
print "\n\nBEST TEST STATISTICS - SENTENCE RESULTS\n";
print "==========================================\n\n";
printf "%40s%15s", "Classifier", "Best for #";
print "     Best for words\n";
print "=" x 100 . "\n";

# Call printTestStats with sentence context
printTestStats(1);

# Globally
print "\n\nBEST TEST STATISTICS - GLOBAL RESULTS\n";
print "========================================\n\n";
printf "%40s%15s", "Classifier", "Best for #";
print "     Best for words\n";
print "=" x 100 . "\n";

# Call printTestStats with global context
printTestStats(2);


# Now we print the worst case analysis

# Before that clear the global data structures that we have to reuse
%teststats = ();
%classifierstats = ();

print "\n\nWORST CASE CLASSIFIERS\n";
print "========================\n\n";

print "FOLLOWING ARE THE WORST CLASSIFIERS:\n";
print "====================================\n\n";

print "ABSTRACT: The worst classifiers (and top $tolerance% above them) are:\n";
print "===================================================================\n\n";

printf "%20s%60s\n", "Word", "Worst classifiers";
print "=" x 100 . "\n";

# Call findWorst with abstract context
findWorst(0);

print "SENTENCE: The worst classifiers (and top $tolerance% above them) are:\n";
print "===================================================================\n\n";

printf "%20s%60s\n", "Word", "Worst classifiers";
print "=" x 100 . "\n";

# Call findWorst with sentence context
findWorst(1);

# Now we print a summary
print "\n\nWORST CASE SUMMARY\n";
print "====================\n\n";

# For abstract
print "WORST CLASSIFIER STATISTICS - ABSTRACT RESULTS (ONLY TOP $usetopx WORST COUNTED)\n";
print "================================================================================\n";
printf "%40s%15s", "Classifier", "Worst for #";
print "     Worst for words\n";
print "=" x 100 . "\n";

# Call printClassStats with abstract context
printClassStats(0);

# For sentence
print "\n\n\nWORST CLASSIFIER STATISTICS - SENTENCE RESULTS (ONLY TOP $usetopx WORST COUNTED)\n";
print "======================================================================================\n";
printf "%40s%15s", "Classifier", "Worst for #";
print "     Worst for words\n";
print "=" x 100 . "\n";

# Call printClassStats with sentence context
printClassStats(1);

# Globally
print "WORST CLASSIFIER STATISTICS - GLOBAL RESULTS (ONLY TOP $usetopx WORST COUNTED)\n";
print "=====================================================================\n";
printf "%40s%15s", "Classifier", "Worst for #";
print "     Worst for words\n";
print "=" x 100 . "\n";

# Call printClassStats with global context
printClassStats(2);

print "\n\n\n";

# Now we want to print the test level statistics

# For abstract
print "\n\nWORST TEST STATISTICS - ABSTRACT RESULTS\n";
print "===========================================\n\n";
printf "%40s%15s", "Classifier", "Worst for #";
print "     Worst for words\n";
print "=" x 100 . "\n";

# Call printTestStats with abstract context
printTestStats(0);

# For sentence
print "\n\nWORST TEST STATISTICS - SENTENCE RESULTS\n";
print "==========================================\n\n";
printf "%40s%15s", "Classifier", "Worst for #";
print "     Worst for words\n";
print "=" x 100 . "\n";

# Call printTestStats with sentence context
printTestStats(1);

print "\n\nWORST TEST STATISTICS - GLOBAL RESULTS\n";
print "========================================\n\n";
printf "%40s%15s", "Classifier", "Worst for #";
print "     Worst for words\n";
print "=" x 100 . "\n";

# Call printTestStats with global context
printTestStats(2);

# Now we have to print the overall accuracy figures
print "\n\nOVERALL ACCURACY FIGURES\n";
print "============================\n\n";

printf "%20s%40s%15s\n", "Context", "Classifier",  "Accuracy";
printf "=" x 75 . "\n";

my $row = sprintf "%20s", "Abstract";
printPerf(0);

$row = sprintf "%20s", "Sentence";
printPerf(1);

$row = sprintf "%20s", "Global";
printPerf(2);

print "\n";

########################################################################
# SUB-ROUTINE SECTION
########################################################################


# Sub procedure to read results from a Weka output file
# and fill them in the results hash
sub readResults {

    # Get the file name and the context to use - abstract or sentence
    my $file = shift @_;
    my $context = shift @_;

    print STDERR "Working on results file $file\n";
    open (HFILE, $file) 
        or die "Failed to open abstract result file $file";

    # $line holds each line that we read
    # $done indicates whether we were successful in
    # finding the Cross Validation results
    my $line;
    my $done = 0;

    # Read each line
    while ($line = <HFILE>) {

        # Check if we find the beginning of the stratified
        # cross validation section
        if ($line =~ /.*Stratified cross.*/) {

            # Skip two lines after that line
            $line = <HFILE>;
            $line = <HFILE>;

            # Then comes the results line, get it
            # and split it on whitespace
            my @tokens = split / +/, $line;

            # Since the NLM data contains 100 instances
            # for each word, correctly classified instances
            # is the same as accuracy percentage, get that
            # and store in results hash
            
            # The results hash contains the outcome of each
            # test for each word and is three level hash
            # For each word, for each test, key element 0
            # in third level indicates results for abstract
            # and key element 1 indicates sentence
            # We have this in the $context variable
            # Update them
            $results{$word_dir}{$test_dir}{$context} = $tokens[3];

            # We found the X valication results, we are done
            $done = 1;

            # Well, not quite! We also need the precision
            # and recall results for each of the senses.
            # Continue reading the file further
            while ($line = <HFILE>) {

                # Search for the detailed results
                if ($line =~ /.*Detailed Accuracy.*/) {

                    # Skip two lines after that
                    $line = <HFILE>;
                    $line = <HFILE>;

                    # Read from the next line onwards upto the
                    # next empty line
                    $line = <HFILE>;
                    chomp $line;
                    
                    while ($line ne "") {

                        @tokens = split(/ +/, $line);

                        $results{$word_dir}{$test_dir}{$tokens[6]}{"PR$context"}
                        = sprintf("PR=%4.3f", $tokens[3]);
                        
                        $results{$word_dir}{$test_dir}{$tokens[6]}{"RE$context"}
                        = sprintf("RE=%4.3f", $tokens[4]);
                        
                        $results{$word_dir}{$test_dir}{$tokens[6]}{"FM$context"}
                        = sprintf("FM=%4.3f", $tokens[5]);

                        $line = <HFILE>;
                        chomp $line;
                    }

                    # Now we are done with reading the precision, recall
                    # and F-measure values too.
                    $done = 2;

                    last;
                }
            }

            if ($done < 2) {

                readSenses($file, $context);
            }

            last;
        }
    }

    if (!$done) {

        $results{$word_dir}{$test_dir}{$context} = -1;

        readSenses($file, $context);
    }
    
    close HFILE;

    return 0;
}

# This functions helps in reading the set of senses for a word
# for which the WEKA analysis of precision, recall and F-measure
# is not available
sub readSenses {

    my $file = shift @_;
    my $context = shift @_;

    # Operate only if we have a file name that we expect
    unless ($file =~ /.*wekaoutput/) {

        return -1;
    }

    # Remove the ".wekaoutput" at the end
    $file =~ s/\.wekaoutput//;

    my $sense;
    my $line;
    my %senses;

    open (HSVAL2, $file) 
        or die "Failed to open ARFF file to find sense classes.";

    while ($line = <HSVAL2>) {

        chomp $line;
        if ($line =~ /.*senseid="(.*)"/) {

            $senses{$1} = 1;
        }
    }
    
    close HSVAL2;
            
    foreach $sense (keys %senses) {

        $results{$word_dir}{$test_dir}{$sense}{"PR$context"} 
        = sprintf("PR=%4.3f", 0);
        $results{$word_dir}{$test_dir}{$sense}{"RE$context"} 
        = sprintf("RE=%4.3f", 0);
        $results{$word_dir}{$test_dir}{$sense}{"FM$context"} 
        = sprintf("FM=%4.3f", 0);
    }

    return 0;
}

# This procedure prints out the detailed accuracy, precision, recall
# and F-measure results for each test for each word
sub printResults {

    my $context = shift @_;

    my $row;
    my $word;
    my $tests;
    my $test;
    my %prfhash;
    my $sense;

    foreach $word (sort keys %results) {

        $tests = $results{$word};

        $row = sprintf("%20s", $word);

        foreach $test (sort keys %{$tests}) {

            $row .= sprintf("%10s", $tests->{$test}{$context});

            if (keys %{$tests->{$test}} > 2) {

                foreach $sense (sort keys %{$tests->{$test}}) {

                    if ($sense ne "0" && $sense ne "1") {

                        if (exists $prfhash{$sense}{"PR"}) {

                            $prfhash{$sense}{"PR"} .= sprintf("%10s",
                                $tests->{$test}{$sense}{"PR$context"});

                            $prfhash{$sense}{"RE"} .= sprintf("%10s",
                                $tests->{$test}{$sense}{"RE$context"});

                            $prfhash{$sense}{"FM"} .= sprintf("%10s",
                                $tests->{$test}{$sense}{"FM$context"});

                        } else {
                            
                            $prfhash{$sense}{"PR"} = sprintf("%20s", $sense);
                            $prfhash{$sense}{"PR"} .= sprintf("%10s",
                                $tests->{$test}{$sense}{"PR$context"});

                            $prfhash{$sense}{"RE"} = sprintf("%20s", "");
                            $prfhash{$sense}{"RE"} .= sprintf("%10s",
                                $tests->{$test}{$sense}{"RE$context"});

                            $prfhash{$sense}{"FM"} = sprintf("%20s", "");
                            $prfhash{$sense}{"FM"} .= sprintf("%10s",
                                $tests->{$test}{$sense}{"FM$context"});
                        }
                    }
                }
            }
        }

        print "$row\n";

        foreach $sense (sort keys %prfhash) {

            print $prfhash{$sense}{"PR"} . "\n";
            print $prfhash{$sense}{"RE"} . "\n";
            print $prfhash{$sense}{"FM"} . "\n";
        }

        print ("\n");

        %prfhash = ();
    }
}

# This procedure finds the best classifiers and gathers statistics
# for them, seperately for abstract, sentence and globally

# IMPORTANT: A very critical assumption in this analysis is that the
# results of the majority classifier are in the directory "test0".
# Without this, the results will not be accurate OR worse - the script
# may fail!

sub findBest {

    my $context = shift @_;

    my @testsarray;
    my @results;

    my $tests;
    my $test;
    my $word;

    # Start processing each word
    foreach $word (sort keys %results) {

        printf "%20s\n", $word;

        $tests = $results{$word};

        # Get the results in descending order
        foreach $test (sort {$tests->{$b}{$context} <=> 
                        $tests->{$a}{$context}} keys %{$tests}) {

            push (@testsarray, $test);
            push (@results, $tests->{$test}{$context});
        }

        # Analyse if a word is useful or not
        my $result;
        my $usefulword = 0;

        if ($context == 0) {
            foreach $result (@results) {

                if ($result >= $tests->{"test0"}{$context}+$threshold) {

                    $usefulword = 1;
                    last;
                }
            }
        } else {

            # We are in sentence context and we should treat a word
            # useful if we find that it was significant in the
            # abstract analysis
            my $testword;

            # Initially assume that the word is useful
            $usefulword = 1;

            # Reject the word if it was rejected in abstract context
            foreach $testword (@words_skipped) {

                if ($word eq $testword) {

                    $usefulword = 0;
                    last;
                }
            }
        }

        # Skip processing if not useful
        if (!$usefulword) {

            @testsarray = ();
            @results = ();

            if ($context == 0) {
                push (@words_skipped, $word);
            }

            next;
        }

        # Find the best classifier accuracy 
        my $maxval = $results[0];
        my $index = 0;

        # Print out those classifiers which are within the accuracy
        # tolerance limit
        while ($index < @results && $results[$index] >= $maxval-$tolerance && 
                $results[$index] >= $tests->{"test0"}{$context}+$threshold) {

            printf "%20s%60s\n", "", "$testtoclassifier{$testsarray[$index]} " . 
                "($testsarray[$index]: $results[$index])";
            $index++;
        }

        print "\n";

        # Print out the accuracy of the first classifier that missed
        # the tolerance limit
        if ($index < @results) {
            printf "%20s%60s\n", "", "First miss: " . 
                "$testtoclassifier{$testsarray[$index]} ($testsarray[$index]:" .
                " $results[$index])";
        }

        # Finally, print the accuracy value of the majority classifier
        printf "%20s%60s\n", "", "Majority: $tests->{\"test0\"}{$context}";


        # Now we are updating the performance related numbers
        # We exclude the two words for which SMO fails since
        # all the instances have the same class - i.e. all instances
        # have the same sense
        if ($word ne "association" && $word ne "fluid") {

            $index = 0;
            while ($index < @results) {

                # Here we update the number for the requested context
                # 0 = abstract
                # 1 = sentence
                if (exists $perfstats{$context}{$testtoclassifier{$testsarray[$index]}}) {

                    $perfstats{$context}{$testtoclassifier{$testsarray[$index]}}{"count"}++;
                    $perfstats{$context}{$testtoclassifier{$testsarray[$index]}}{"value"} += $results[$index];


                } else {

                    $perfstats{$context}{$testtoclassifier{$testsarray[$index]}}{"count"} = 1;
                    $perfstats{$context}{$testtoclassifier{$testsarray[$index]}}{"value"} = $results[$index];
                }

                # Now we update the global performance numbers
                # context = 2
                if (exists $perfstats{2}{$testtoclassifier{$testsarray[$index]}}) {

                    $perfstats{2}{$testtoclassifier{$testsarray[$index]}}{"count"}++;
                    $perfstats{2}{$testtoclassifier{$testsarray[$index]}}{"value"} += $results[$index];


                } else {

                    $perfstats{2}{$testtoclassifier{$testsarray[$index]}}{"count"} = 1;
                    $perfstats{2}{$testtoclassifier{$testsarray[$index]}}{"value"} = $results[$index];
                }

                $index++;
            }
        }

        # Now we want to find out the the statistics at the level of each
        # individual test. This loop also takes care of updating the
        # temporary %classifierstatstemp hash which is used for finding
        # the classifier level statistics

        # We print the number of word senses for each word which was
        # best analyzed by a classifier. Get this from the output of majority
        # class
        my $wordsensecount = (keys %{$tests->{"test0"}}) - 2;

        # Start the actual loop that does the analysis for test level
        # and classifier level statistics
        $index = 0;
        while ($index < @results && $results[$index] >= $maxval-$usetopx+1 
            && $results[$index] >= $maxval-$tolerance && 
            $results[$index] >= $tests->{"test0"}{0}+$threshold) {

            # This first if..else handles updating the stats for the
            # specific requested context
            if (exists $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}) {
                
                $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"count"}++;
                $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"words"} .= "$word($wordsensecount) ";

            } else {

                $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"count"} = 1;
                $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"words"} = "$word($wordsensecount) ";

            }

            # This if..else handles updating the stats at global level
            if (exists $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}) {
                
                $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"count"}++;
                $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"words"} .= "$word($wordsensecount) ";

            } else {

                $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"count"} = 1;
                $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"words"} = "$word($wordsensecount) ";

            }

            # Then we simply record that this particular classifier was
            # the best for some set of options. Note that the assignment
            # statement below is intentional - it counts a particular
            # classifier as best only once - even if it was best for the
            # word with different set of options in different test
            # cases
            $classifierstatstemp{"$testtoclassifier{$testsarray[$index]}"} = 1;
            $index++;
        }

        # Now we actually update the count at classifier level. We do
        # not want to do it inside the above loop because - we increase
        # it only once per classifier - not for its different options
        my $classifier;
        foreach $classifier (keys %classifierstatstemp) {

            # First do it for the requested context
            if (exists $classifierstats{$context}{$classifier}) {

                $classifierstats{$context}{$classifier}{"count"}++;
                $classifierstats{$context}{$classifier}{"words"} .= "$word($wordsensecount) ";

            } else {

                $classifierstats{$context}{$classifier}{"count"} = 1;
                $classifierstats{$context}{$classifier}{"words"} = "$word($wordsensecount) ";

            }

            # Then do it at global level
            if (exists $classifierstats{2}{$classifier}) {

                $classifierstats{2}{$classifier}{"count"}++;
                $classifierstats{2}{$classifier}{"words"} .= "$word($wordsensecount) ";

            } else {

                $classifierstats{2}{$classifier}{"count"} = 1;
                $classifierstats{2}{$classifier}{"words"} = "$word($wordsensecount) ";

            }
        }

        # Clear all temporary variables that we have used
        %classifierstatstemp = ();

        @testsarray = ();
        @results = ();

        print "\n\n";
    }

    # Print the words that we have skipped for this analysis
    # These are the ones that did not perform good enough above the
    # majority classifier
    print "The words that were skipped from analysis were: @words_skipped\n\n";
    print "\n";
}

# This procedure simply loops over the %classifierstats hash printing
# out the statistics for the requested context
sub printClassStats {

    my $context = shift @_;

    my $classifier;

    foreach $classifier (sort { $classifierstats{$context}{$b}{"count"} <=> $classifierstats{$context}{$a}{"count"} } keys %{$classifierstats{$context}}) {

        printf "%40s%15s", $classifier, $classifierstats{$context}{$classifier}{"count"};
        print "     $classifierstats{$context}{$classifier}{\"words\"}\n";
    }
}

# This procedure simply loops over the %teststats hash printing
# out the statistics for the requested context
sub printTestStats {

    my $context = shift @_;

    my $classifier;

    foreach $classifier (sort { $teststats{$context}{$b}{"count"} <=> $teststats{$context}{$a}{"count"} } keys %{$teststats{$context}}) {

        printf "%40s%15s", $classifier, $teststats{$context}{$classifier}{"count"};
        print "     $teststats{$context}{$classifier}{\"words\"}\n";
    }
}


# This procedure finds the worst classifiers and gathers statistics
# for them, seperately for abstract, sentence and globally

# IMPORTANT: A very critical assumption in this analysis is that the
# results of the majority classifier are in the directory "test0".
# Without this, the results will not be accurate OR worse - the script
# may fail!

sub findWorst {

    my $context = shift @_;

    my @testsarray;
    my @results;

    my $tests;
    my $test;
    my $word;

    # Start processing each word
    foreach $word (sort keys %results) {

        printf "%20s\n", $word;

        $tests = $results{$word};

        # Get the results in ascending order
        foreach $test (sort {$tests->{$a}{$context} <=> 
                        $tests->{$b}{$context}} keys %{$tests}) {

            push (@testsarray, $test);
            push (@results, $tests->{$test}{$context});
        }

        # Find the worst classifier accuracy 
        my $minval = $results[0];
        my $index = 0;

        # Print out those classifiers which are within the accuracy
        # tolerance limit
        while ($index < @results && $results[$index] <= $minval+$tolerance) {

            printf "%20s%60s\n", "", "$testtoclassifier{$testsarray[$index]} " . 
                "($testsarray[$index]: $results[$index])";
            $index++;
        }

        print "\n";

        # Print out the accuracy of the first classifier that missed
        # the tolerance limit
        if ($index < @results) {
            printf "%20s%60s\n", "", "First miss: " . 
                "$testtoclassifier{$testsarray[$index]} ($testsarray[$index]:" .
                " $results[$index])";
        }

        # Finally, print the accuracy value of the majority classifier
        printf "%20s%60s\n", "", "Majority: $tests->{\"test0\"}{$context}";

        # Now we want to find out the the statistics at the level of each
        # individual test. This loop also takes care of updating the
        # temporary %classifierstatstemp hash which is used for finding
        # the classifier level statistics

        # We print the number of word senses for each word which was
        # best analyzed by a classifier. Get this from the output of majority
        # class
        my $wordsensecount = (keys %{$tests->{"test0"}}) - 2;

        # Start the actual loop that does the analysis for test level
        # and classifier level statistics
        $index = 0;
        while ($index < @results && $results[$index] <= $minval+$usetopx-1 
            && $results[$index] <= $minval+$tolerance) {

            # This first if..else handles updating the stats for the
            # specific requested context
            if (exists $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}) {
                
                $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"count"}++;
                $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"words"} .= "$word($wordsensecount) ";

            } else {

                $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"count"} = 1;
                $teststats{$context}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"words"} = "$word($wordsensecount) ";

            }

            # This if..else handles updating the stats at global level
            if (exists $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}) {
                
                $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"count"}++;
                $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"words"} .= "$word($wordsensecount) ";

            } else {

                $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"count"} = 1;
                $teststats{2}{"$testtoclassifier{$testsarray[$index]}:$testsarray[$index]"}{"words"} = "$word($wordsensecount) ";

            }

            # Then we simply record that this particular classifier was
            # the worst for some set of options. Note that the assignment
            # statement below is intentional - it counts a particular
            # classifier as worst only once - even if it was worst for the
            # word with different set of options in different test
            # cases
            $classifierstatstemp{"$testtoclassifier{$testsarray[$index]}"} = 1;
            $index++;
        }

        # Now we actually update the count at classifier level. We do
        # not want to do it inside the above loop because - we increase
        # it only once per classifier - not for its different options
        my $classifier;
        foreach $classifier (keys %classifierstatstemp) {

            # First do it for the requested context
            if (exists $classifierstats{$context}{$classifier}) {

                $classifierstats{$context}{$classifier}{"count"}++;
                $classifierstats{$context}{$classifier}{"words"} .= "$word($wordsensecount) ";

            } else {

                $classifierstats{$context}{$classifier}{"count"} = 1;
                $classifierstats{$context}{$classifier}{"words"} = "$word($wordsensecount) ";

            }

            # Then do it at global level
            if (exists $classifierstats{2}{$classifier}) {

                $classifierstats{2}{$classifier}{"count"}++;
                $classifierstats{2}{$classifier}{"words"} .= "$word($wordsensecount) ";

            } else {

                $classifierstats{2}{$classifier}{"count"} = 1;
                $classifierstats{2}{$classifier}{"words"} = "$word($wordsensecount) ";

            }
        }

        # Clear all temporary variables that we have used
        %classifierstatstemp = ();

        @testsarray = ();
        @results = ();

        print "\n\n";
    }
}

# This procedure simply prints the accuracy numbers for a given context
sub printPerf {

    my $context = shift @_;

    my $classifier;

    foreach $classifier (keys %{$perfstats{$context}}) {

        $row .= sprintf "%40s%10.2f(%3d)\n", $classifier, $perfstats{$context}{$classifier}{"value"} / $perfstats{$context}{$classifier}{"count"}, $perfstats{$context}{$classifier}{"count"}; 
        $row .= sprintf "%20s", "";
    }

    print $row . "\n";
}

